; ==================================================================================================
; Title:   COM_Dispatch.inc
; Author:  G.Friedrich
; Version: 1.0.0
; Purpose: COM dispatching helper procedures compatible with ObjAsm32 definitions.
; Notes:   Version 1.0.0, January 2006
;           - First release. Source code hints taken from Japheth's DispHelp.inc.
; ==================================================================================================


externdef IID_NULL:GUID

ExecDispatch proto C, :Pointer, :dword, :dword, :dword, :dword, :vararg
CallDispatch proto :Pointer, :dword, :dword, :dword, :Pointer, :Pointer

; 
; Macro:     GetMethodDispatchID
; Purpose:   Returns the method ID of a given interface.
; Arguments: Arg1: Instance Pointer.
;            Arg2: Interface Name.
;            Arg3: Method Name.
;            Arg4: Destrination variable.
;            Arg5: (optional) Local Identifier.
; Usage:     GetMethodDispatchID pIDemo16, IDemo16, "Increment", dDispatchID

GetMethodDispatchID macro InterfacePointer:req, InterfaceName:req, MethodName:req, DstVar:req, LCID
    local ArgLCID, ArgDest

    ifb <LCID>
      invoke GetUserDefaultLCID
      lea ecx, DstVar                   ;;use ecx as address of destination variable
      ArgDest textequ <ecx>
      ArgLCID textequ <eax>             ;;use eax as LCID value
    else
      ifidni <LCID>, <ecx>
        lea eax, DstVar                 ;;use eax as address of destination variable
        ArgDest textequ <eax>
        ArgLCID textequ <ecx>           ;;use ecx as LCID value
      else                              ;;ecx was not used here => use it!
        lea ecx, DstVar                 ;;use ecx as address of destination variable
        ArgDest textequ <ecx>
        ArgLCID textequ <LCID>          ;;use LCID as given in the parameter list
      endif
    endif
    push $OfsCUStr(MethodName)          ;;Allocate method name and store its address onto stack
    ICall InterfacePointer::InterfaceName.GetIDsOfNames, offset IID_NULL, addr [esp + 12], 1, ArgLCID, ArgDest
    add esp, 4                          ;;Restore stack
endm

; 
; Macro:     DEFINE_DISPMETHOD
; Purpose:   Defines some equates to use a Dispatch interface.
; Arguments: Arg1: Interface name.
;            Arg2: Method name.
;            Arg3: Method Dispatch ID.
;            Arg4: Method type.
;            Arg5: Return value type.
;            Arg6: Method argument types.
; Returns:   Nothing

DEFINE_DISPMETHOD macro IName, MName:req, DispID:req, MType:req, RetType, Args:vararg
    local i, Parm, ArgTypes, ReturnType

    ArgTypes textequ <>
    i = 0
    for Parm, <Args>
      i = i + 1
      ArgTypes CatStr ArgTypes, %Parm, <,>
    endm
    if i gt 0
      ArgTypes SubStr ArgTypes, 1, @SizeStr(%ArgTypes) - 1    ;;Remove last comma
    endif

    ifnb <RetType>
      ReturnType textequ %RetType
    else
      ReturnType textequ <-1>           ;;Set Flag if no returns value is required
    endif

    DISPATCH_INFO_&IName&_&MName& CatStr <DispID>, <,>, %DISPATCH_&MType&, <,>, %i
    DISPATCH_TYPE_&IName&_&MName& CatStr ReturnType, <,>, ArgTypes
endm

; 
; Macro:     AutoDispatch
; Purpose:   Performs a COM method call using the Dispatch interface. Variant arguments are created
;            automatically onto stack.
; Arguments: Arg1: pInterface::Interface.Method
;            Arg2: Method arguments
; Returns:   eax = Method return value.
;            ecx = Argument index if an error happened.

AutoDispatch macro Cast:req, Args:vararg
    local we, wa, PName, IName, MName

    we InStr <Cast>, <::>
    if we ne 0
      PName SubStr <Cast>, 1, we - 1
      wa InStr we + 2, <Cast>, <.>
      if wa ne 0
        IName SubStr <Cast>, we + 2, wa - we - 2
        MName SubStr <Cast>, wa + 1
      else
        .err <DispatchArgs format error. Dot-method missing.>
        exitm
      endif
    else
      .err <DispatchArgs format error. Double points missing.>
      exitm
    endif

    invoke ExecDispatch, PName, @CatStr(<DISPATCH_INFO_>, %IName, <_>, %MName), \
                         @CatStr(<DISPATCH_TYPE_>, %IName, <_>, %MName), Args
endm

; 
; Macro:     Dispatch
; Purpose:   Performs a COM method call using the Dispatch interface. Varint arguments have to
;            be supplied.
; Arguments: Arg1: pInterface::Interface.Method
;            Arg2: -> Method Variant arguments.
;            Arg3: -> Return Variant.
; Returns:   eax = Method return value.
;            ecx = Argument index if an error happened.

Dispatch macro Cast:req, pVarArgs:req, pVarRet:=<NULL>
    local we, wa, PName, IName, MName

    we InStr <Cast>, <::>
    if we ne 0
      PName SubStr <Cast>, 1, we - 1
      wa InStr we + 2, <Cast>, <.>
      if wa ne 0
        IName SubStr <Cast>, we + 2, wa - we - 2
        MName SubStr <Cast>, wa + 1
      else
        .err <Dispatch format error. Dot-method missing.>
        exitm
      endif
    else
      .err <Dispatch format error. Double points missing.>
      exitm
    endif

    invoke CallDispatch, PName, @CatStr(<DISPATCH_INFO_>, %IName, <_>, %MName), pVarArgs, pVarRet
endm
